@@ -1,14 +1,17 @@
 import logging
 import os
+import re
 import string
 import struct
 import zlib
+import glob
 from collections import OrderedDict
 from pathlib import Path
 
 import vdf
 
-from .util import lower_dict, is_steam_deck
+from ._vdf import binary_loads as vendored_binary_loads
+from .util import is_steam_deck, lower_dict
 
 __all__ = (
     "COMMON_STEAM_DIRS", "SteamApp", "find_steam_installations",
@@ -27,6 +30,13 @@ COMMON_STEAM_DIRS = [
     ".local/share/Steam",
 ]
 
+SNAP_STEAM_DIRS = [
+    "snap/steam/common/.local/share/Steam",
+
+    # Location if 'experimental.hidden-snap-folder' Snap feature flag is set
+    ".snap/data/steam/common/.local/share/Steam",
+]
+
 OS_RELEASE_PATHS = [
     "/run/host/os-release",  # The host file if we're inside a Flatpak sandbox
     "/etc/os-release"
@@ -255,8 +265,31 @@ class SteamApp(object):
 
         # If Steam path was provided, also populate the icon path
         if steam_path:
-            icon_path = \
-                steam_path / "appcache" / "librarycache" / f"{appid}_icon.jpg"
+            # The icon path might be located in one of the two locations:
+            # 1. `<steam_path>/appcache/librarycache/<appid>/<40 hex chars>.jpg
+            # 2. `<steam_path>/appcache/librarycache/<appid>_icon.jpg`
+            #
+            # There doesn't appear to be any other way to determine which is
+            # used other than by checking. This incurs some I/O for each app.
+            icon_path = None
+            library_cache_path = steam_path / "appcache" / "librarycache"
+            try:
+                # Try 1st location. This appears to be the newest, so this
+                # should hopefully be the first match, at least on newer Steam
+                # installations.
+                app_lib_cache_path = library_cache_path / str(appid)
+
+                icon_path = next(
+                    path for path in app_lib_cache_path.iterdir()
+                    if re.match(r"[a-f0-9]{40}\.jpg", path.name)
+                )
+            except (StopIteration, FileNotFoundError):
+                # Try 2nd location
+                icon_path = library_cache_path / f"{appid}_icon.jpg"
+
+                if not icon_path.is_file():
+                    # No icon was found
+                    icon_path = None
 
         # Check if the app requires another app. This is the case with
         # newer versions of Proton, which use Steam Runtimes installed as
@@ -388,14 +421,19 @@ def find_steam_installations():
 
             candidates[(str(steam_path), str(steam_root_))] = True
 
-    # Check for Flatpak Steam separately and ensure we don't mix steam_root
-    # and steam_path from Flatpak and non-Flatpak installations of Steam.
+    # Check for Flatpak and Snap Steam separately and ensure we don't mix steam_root
+    # and steam_path from Flatpak/Snap and native installations of Steam.
     steam_path = \
         Path.home() / ".var/app/com.valvesoftware.Steam/data/Steam"
     steam_path = steam_path.resolve()
     if has_steamapps_dir(steam_path):
         candidates[(str(steam_path), str(steam_path))] = True
 
+    for steam_dir in SNAP_STEAM_DIRS:
+        steam_path = (Path.home() / steam_dir).resolve()
+        if has_steamapps_dir(steam_path):
+            candidates[(str(steam_path), str(steam_path))] = True
+
     for steam_path, _ in candidates.keys():
         logger.info("Found Steam directory at %s", steam_path)
 
@@ -471,6 +509,7 @@ def find_legacy_steam_runtime_path(steam_root):
 
 APPINFO_STRUCT_HEADER = "<4sL"
 APPINFO_V28_STRUCT_SECTION = "<LLLLQ20sL20s"
+APPINFO_V29_STRUCT_SECTION = "<LLLLQ20sL20s"
 
 
 def iter_appinfo_sections(path):
@@ -506,6 +545,69 @@ def iter_appinfo_sections(path):
             if i == len(data) - 4:
                 return
 
+    def _iter_v29_appinfo(data, start):
+        """
+        Parse and iterate appinfo.vdf version 29.
+        """
+        i = start
+
+        # The header contains the offset to the key table
+        key_table_offset = struct.unpack("<q", data[i:i+8])[0]
+        key_table = []
+
+        key_count = struct.unpack(
+            "<i", data[key_table_offset:key_table_offset+4]
+        )[0]
+
+        table_i = key_table_offset + 4
+        for _ in range(0, key_count):
+            key = bytearray()
+            while True:
+                key.append(data[table_i])
+                table_i += 1
+
+                if key[-1] == 0:
+                    key_table.append(
+                        key[0:-1].decode("utf-8", errors="replace")
+                    )
+                    break
+
+        i += 8
+
+        section_size = struct.calcsize(APPINFO_V29_STRUCT_SECTION)
+        while True:
+            # We don't need any of the fields besides 'entry_size',
+            # which is used to determine the length of the variable-length VDF
+            # field.
+            # Still, here they are for posterity's sake.
+            (appid, entry_size, infostate, last_updated, access_token,
+             sha_hash, change_number, vdf_sha_hash) = struct.unpack(
+                APPINFO_V29_STRUCT_SECTION, data[i:i+section_size])
+            vdf_section_size = entry_size - (section_size - 8)
+
+            i += section_size
+
+            try:
+                vdf_d = vdf.binary_loads(
+                    data[i:i+vdf_section_size], key_table=key_table
+                )
+            except TypeError:
+                # System 'vdf' is too old and does not support 'key_table',
+                # use the bundled one instead. This is cursed, but it's
+                # so far the only reasonable option without a proper maintained
+                # release on PyPI.
+                vdf_d = vendored_binary_loads(
+                    data[i:i+vdf_section_size], key_table=key_table
+                )
+
+            vdf_d = lower_dict(vdf_d)
+            yield vdf_d
+
+            i += vdf_section_size
+
+            if i == key_table_offset - 4:
+                return
+
     logger.debug("Loading appinfo.vdf in %s", path)
 
     # appinfo.vdf is not actually a (binary) VDF file, but a binary file
@@ -527,6 +629,8 @@ def iter_appinfo_sections(path):
 
     if magic == b'(DV\x07':
         yield from _iter_v28_appinfo(data, i)
+    elif magic == b')DV\x07':
+        yield from _iter_v29_appinfo(data, i)
     else:
         raise SyntaxError(
             "Invalid file magic number. The appinfo.vdf version might not be "
@@ -601,7 +705,10 @@ def find_steam_compat_tool_app(steam_path, steam_apps, appid=None):
     # Multiple configuration values might be found. In such case, select the
     # one with the highest priority.
     config_vdf_path = steam_path / "config" / "config.vdf"
-    content = config_vdf_path.read_text()
+    # config.vdf might contain invalid UTF-8 characters in the
+    # `SDL_GamepadBind` field. We don't use that in any way, so we can deal
+    # with the invalid characters by just ignoring them.
+    content = config_vdf_path.read_text(errors="replace")
 
     vdf_data = lower_dict(vdf.loads(content))
 
@@ -728,15 +835,18 @@ def find_steam_compat_tool_app(steam_path, steam_apps, appid=None):
         )
         potential_names.append(tool_name)
 
-    # Get the first name that was valid, or use stable Proton as fallback
+    # Get the first name that was valid,
+    # or use experimental or stable Proton as fallback
     try:
-        compat_tool_name = next(name for name in potential_names if name)
+        compat_tool_names = [
+            next(name for name in potential_names if name)
+        ]
     except StopIteration:
         logger.info(
             "No compatibility tool found by reading Steam configuration. "
-            "Using stable version of Proton as fallback."
+            "Using Proton Experimental or stable Proton as fallback."
         )
-        compat_tool_name = "proton-stable"
+        compat_tool_names = ["proton-experimental", "proton-stable"]
 
     # We've got a compatibility tool name,
     # now there are two possible ways to find the installation
@@ -746,30 +856,31 @@ def find_steam_compat_tool_app(steam_path, steam_apps, appid=None):
     #    to parse a binary configuration file to find the App ID
 
     # Let's try option 1 first
-    try:
-        app = next(
-            app for app in steam_apps if app.name == compat_tool_name
-        )
-        logger.info(
-            "Found active custom compatibility tool: %s", app.name
-        )
-        return app
-    except StopIteration:
-        pass
+    for compat_tool_name in compat_tool_names:
+        try:
+            app = next(
+                app for app in steam_apps if app.name == compat_tool_name
+            )
+            logger.info(
+                "Found active custom compatibility tool: %s", app.name
+            )
+            return app
+        except StopIteration:
+            pass
 
-    # Try option 2:
-    # Find the corresponding App ID from <steam_path>/appcache/appinfo.vdf
-    tool_app = _get_tool_app(
-        compat_tool_name=compat_tool_name,
-        steam_apps=steam_apps,
-        steam_play_manifest=steam_play_manifest
-    )
-
-    if tool_app:
-        logger.info(
-            "Found active compatibility tool: %s", tool_app.name
+        # Try option 2:
+        # Find the corresponding App ID from <steam_path>/appcache/appinfo.vdf
+        tool_app = _get_tool_app(
+            compat_tool_name=compat_tool_name,
+            steam_apps=steam_apps,
+            steam_play_manifest=steam_play_manifest
         )
-        return tool_app
+
+        if tool_app:
+            logger.info(
+                "Found active compatibility tool: %s", tool_app.name
+            )
+            return tool_app
 
     logger.error("Could not find configured Proton installation!")
 
@@ -873,6 +984,93 @@ def get_steam_lib_paths(steam_path):
     Return a list of any Steam directories including any user-added
     Steam library folders
     """
+    def _case_insensitive_glob_pattern(pattern):
+        """
+        Convert glob pattern into a case-insensitive pattern
+        """
+        return "".join(
+            f"[{char_.upper()}{char_.lower()}]" if char_.isalpha()
+            else glob.escape(char_)
+            for char_ in pattern
+        )
+
+    def resolve_library_folder(path):
+        """
+        Resolve the Steam library folder for the given path found in
+        `libraryfolders.vdf` file. The final path is located in the
+        case-insensitive `steamapps` directory.
+        """
+        xdg_steam_path = Path.home() / ".local/share/Steam"
+        flatpak_steam_path = \
+            Path.home() / ".var/app/com.valvesoftware.Steam/data/Steam"
+
+        is_library_folder_xdg_steam = str(path) == str(xdg_steam_path)
+        is_flatpak_steam = str(steam_path) == str(flatpak_steam_path)
+
+        # Adjust the path if the library folder is "~/.local/share/Steam"
+        # and we're looking for library folders in
+        # "~/.var/app/com.valvesoftware.Steam"
+        # This is because ~/.local/share/Steam inside a Steam Flatpak
+        # sandbox corresponds to a different location, and we need to
+        # adjust for that.
+        if is_library_folder_xdg_steam and is_flatpak_steam:
+            path = flatpak_steam_path
+
+        # Steam matches library folder paths case-insensitively.
+        candidates = [
+            Path(candidate) for candidate
+            in glob.glob(_case_insensitive_glob_pattern(str(path)))
+        ]
+
+        logger.debug(
+            "Following candidates found for the Steam library path %s: %s",
+            path, candidates
+        )
+
+        if not candidates:
+            logger.warning(
+                "Steam library folder %s in Steam configuration does not "
+                "exist",
+                path
+            )
+            return None
+
+        # In case of multiple matches, prioritize those that contain
+        # a 'steamapps' subdirectory
+        candidates.sort(
+            key=lambda path: any(
+                path.name.lower() == "steamapps"
+                for path in path.iterdir()
+            ),
+            reverse=True
+        )
+
+        # In case multiple directories match, pick the first one,
+        # log a warning and hope for the best
+        if len(candidates) > 1:
+            logger.warning(
+                "Multiple Steam library folders with name %s were found. "
+                "Selecting %s.",
+                path, candidates[0]
+            )
+
+        if str(candidates[0]) != str(path):
+            logger.warning(
+                "Steam library path %s in configuration differs from "
+                "found path %s. Was this path renamed?",
+                path, candidates[0]
+            )
+
+        path = candidates[0]
+
+        logger.debug(
+            "Found Steam library folder %s. Is Flatpak path: %s, "
+            "Is XDG Steam path: %s.",
+            path, is_flatpak_steam, is_library_folder_xdg_steam
+        )
+
+        return path
+
     def parse_library_folders(data):
         """
         Parse the Steam library folders in the VDF file using the given data
@@ -894,29 +1092,17 @@ def get_steam_lib_paths(steam_path):
                 # and nothing else
                 path = Path(value)
 
-            xdg_steam_path = Path.home() / ".local/share/Steam"
-            flatpak_steam_path = \
-                Path.home() / ".var/app/com.valvesoftware.Steam/data/Steam"
+            try:
+                path = resolve_library_folder(path)
 
-            is_library_folder_xdg_steam = str(path) == str(xdg_steam_path)
-            is_flatpak_steam = str(steam_path) == str(flatpak_steam_path)
-
-            # Adjust the path if the library folder is "~/.local/share/Steam"
-            # and we're looking for library folders in
-            # "~/.var/app/com.valvesoftware.Steam"
-            # This is because ~/.local/share/Steam inside a Steam Flatpak
-            # sandbox corresponds to a different location, and we need to
-            # adjust for that.
-            if is_library_folder_xdg_steam and is_flatpak_steam:
-                path = flatpak_steam_path
-
-            logger.debug(
-                "Found Steam library folder %s. Is Flatpak path: %s, "
-                "Is XDG Steam path: %s.",
-                path, is_flatpak_steam, is_library_folder_xdg_steam
-            )
-
-            library_folders.append(path)
+                if path:
+                    library_folders.append(path)
+            except OSError:
+                logger.warning(
+                    "Could not resolve Steam library folder %s", path,
+                    exc_info=True
+                )
+                continue
 
         logger.info(
             "Found %d Steam library folders", len(library_folders)
@@ -933,16 +1119,27 @@ def get_steam_lib_paths(steam_path):
             "directory. Installation may be incomplete or broken."
         ) from exc
 
+    folders_vdf_content = None
     try:
-        library_folders = parse_library_folders(folders_vdf_path.read_text())
+        folders_vdf_content = folders_vdf_path.read_text()
     except OSError:
+        logger.warning(
+            "Could not read 'libraryfolders.vdf'. "
+            "Maybe no additional Steam library folders haven't been added?",
+            exc_info=True
+        )
         # libraryfolders.vdf doesn't exist; maybe no Steam library folders
         # are set?
         library_folders = []
-    except SyntaxError as exc:
-        raise ValueError(
-            f"Library folder configuration file {folders_vdf_path} is corrupted"
-        ) from exc
+
+    if folders_vdf_content:
+        try:
+            library_folders = parse_library_folders(folders_vdf_content)
+        except SyntaxError as exc:
+            raise ValueError(
+                f"Library folder configuration file {folders_vdf_path} is "
+                f"corrupted"
+            ) from exc
 
     paths = [steam_path] + library_folders
 
@@ -1007,24 +1204,40 @@ def get_custom_compat_tool_installations_in_dir(compat_tool_dir):
             )
             continue
 
+        if not vdf_data:
+            logger.warning(
+                "Compatibility tool declaration at %s is empty. You may need "
+                "to reinstall the application.",
+                vdf_path
+            )
+            continue
+
         # Traverse to 'compatibilitytools/compat_tools' in a case-insensitive
         # way. This is done because we can't turn all keys recursively to
         # lowercase from the get-go; the app name is stored as a key.
-        compat_tools = {k.lower(): v for k, v in vdf_data.items()}
-        compat_tools = compat_tools["compatibilitytools"]
-        compat_tools = {
-            k.lower(): v for k, v in compat_tools.items()
-        }
-        compat_tools = compat_tools["compat_tools"]
-        internal_name = list(compat_tools.keys())[0]
-        tool_info = compat_tools[internal_name]
+        try:
+            compat_tools = {k.lower(): v for k, v in vdf_data.items()}
+            compat_tools = compat_tools["compatibilitytools"]
+            compat_tools = {
+                k.lower(): v for k, v in compat_tools.items()
+            }
+            compat_tools = compat_tools["compat_tools"]
+            internal_name = list(compat_tools.keys())[0]
+            tool_info = compat_tools[internal_name]
 
-        # We can now convert the remainder into lowercase
-        tool_info = lower_dict(tool_info)
+            # We can now convert the remainder into lowercase
+            tool_info = lower_dict(tool_info)
 
-        install_path_name = tool_info["install_path"]
-        from_oslist = tool_info["from_oslist"]
-        to_oslist = tool_info["to_oslist"]
+            install_path_name = tool_info["install_path"]
+            from_oslist = tool_info["from_oslist"]
+            to_oslist = tool_info["to_oslist"]
+        except KeyError:
+            logger.warning(
+                "Compatibility tool declaration at %s is incomplete. You may "
+                "need to reinstall the application",
+                vdf_path
+            )
+            continue
 
         if from_oslist != "windows" or to_oslist != "linux":
             continue
